@@ -3,11 +3,14 @@ from __future__ import annotations #only > 3.7, better to find a different solut
 import os
 from abc import ABC, abstractmethod
 import json
+import itertools
 
 import hashlib
 from urllib.parse import urlparse
 from urllib.parse import parse_qs
 
+from pathlib import PurePosixPath
+
 import google_auth_httplib2
 from googleapiclient.discovery import build
 from google_auth_oauthlib.flow import InstalledAppFlow
@@ -52,6 +55,12 @@ class CheckSumError(Exception):
     pass
 
 class AmbiguousPathError(Exception):
+    def __init__(self, *args, **kwargs):
+        if "duplicates" in kwargs:
+            self.duplicates = kwargs["duplicates"]
+        super().__init__(*args)
+
+class InvalidUrlError(Exception):
     pass
 
 class Credentials(google.oauth2.credentials.Credentials,
@@ -120,7 +129,7 @@ class DriveItem(ABC):
         else:
             return None
 
-    def rename(self, target):
+    def rename(self, target, ignore_existing=False):
         splitpath = target.rsplit('/', 1)
         if len(splitpath) == 1:
             new_name = splitpath[0]
@@ -131,12 +140,23 @@ class DriveItem(ABC):
             if not parent.isfolder():
                 raise NotADirectoryError()
 
-        self.move(parent, new_name)
+        self.move(parent, new_name, ignore_existing=ignore_existing)
 
-    def move(self, new_dest, new_name=None):
+    def move(self, new_dest, new_name=None, ignore_existing=False):
+        new_name = new_name or self.name
+        if not ignore_existing:
+            try:
+                child = new_dest.child(new_name)
+                raise FileExistsError("Filename already exists ({name}).".format(name=child.name))
+            except FileNotFoundError:
+                pass
+            except AmbiguousPathError:
+                raise FileExistsError("Filename already exists multiple times: {name}".format(name=new_name))
+
+        #TODO: Don't use exception for flow control here. Maybe implement exists()
         result = self.drive.service.files().update(
                                 fileId=self.id,
-                                body={"name": new_name or self.name},
+                                body={"name": new_name},
                                 addParents=new_dest.id,
                                 removeParents=self.parent.id,
                                 fields='name, parents',
@@ -188,12 +208,13 @@ class DriveFolder(DriveItem):
         child = next(gen, None)
         if child is None:
             raise FileNotFoundError(name)
-        if next(gen, None) is not None:
-            raise AmbiguousPathError("Two or more files {name}".format(name=name))
+        next_child = next(gen, None)
+        if next_child is not None:
+            duplicates = itertools.chain((child, next_child), gen)
+            raise AmbiguousPathError("Two or more files {name}".format(name=name), duplicates=duplicates)
         return child
         
     def children(self, name=None, folders=True, files=True, trashed=False, pageSize=100, orderBy=None):
-        #TODO: Add "name" argument
         query = "'{this}' in parents".format(this=self.id)
 
         if name:
@@ -213,23 +234,36 @@ class DriveFolder(DriveItem):
 
         return self.drive.items_by_query(query, pageSize=pageSize, orderBy=orderBy, spaces=self.spaces)
 
-    def mkdir(self, name):
-        try:
-            file_ = self.child(name)
-            if not file_.isfolder():
-                raise FileExistsError("Filename already exists ({name}) and it's not a folder.".format(name=name))
-            return file_
-        except FileNotFoundError:
-            #TODO: Don't use exception for flow control here. Maybe implement exists()
-            file_metadata = {
+    def mkdir(self, name, ignore_existing=False):
+        if not ignore_existing:
+            try:
+                file_ = self.child(name)
+                if file_.isfolder():
+                    return file_
+                else:
+                    raise FileExistsError("Filename already exists ({name}) and it's not a folder.".format(name=name))
+            except FileNotFoundError:
+                pass
+            except AmbiguousPathError:
+                raise FileExistsError("Filename already exists multiple times: {name}".format(name=name))
+        #TODO: Don't use exception for flow control here. Maybe implement exists()
+        file_metadata = {
+            'name': name, 
                 'name': name, 
-                'mimeType': 'application/vnd.google-apps.folder',
-                'parents': [self.id]
-            }
-            result = self.drive.service.files().create(body=file_metadata, fields=self.drive.default_fields).execute()
-            return self._reply_to_object(result)
+            'name': name, 
+            'mimeType': 'application/vnd.google-apps.folder',
+            'parents': [self.id]
+        }
+        result = self.drive.service.files().create(body=file_metadata, fields=self.drive.default_fields).execute()
+        return self._reply_to_object(result)
         
-    def new_file(self, filename):
+    def new_file(self, filename, ignore_existing=False):
+        if not ignore_existing:
+            try:
+                self.child(filename)
+                raise FileExistsError("Filename already exists ({name}).".format(name=filename))
+            except FileNotFoundError:
+                pass
         return DriveFile(self.drive, [self.id], filename)
         
     def child_from_path(self, path) -> DriveItem:
@@ -523,6 +557,24 @@ class GoogleDrive(DriveFolder):
             raise NotAuthenticatedError("Could not get requested scopes")
         return Credentials.to_json(creds)
 
+    @classmethod
+    def url_to_id(cls, url: str):
+        url = urlparse(url)
+        if url.netloc != 'drive.google.com':
+            raise InvalidUrlError(url)
+        if url.path in ('/uc', '/open'):
+            try:
+                return parse_qs(url.query)['id'][0]
+            except KeyError:
+                raise InvalidUrlError("Has no ID", url)
+        else:
+            try:
+                path = PurePosixPath(url.path)
+                assert path.parts[0:3] == ('/', 'file','d')
+                return path.parts[3]
+            except (AssertionError, KeyError):
+                raise InvalidUrlError(url)
+
     def __init__(self, creds, autorefresh=True):
         try:
             self.creds = Credentials.from_json(creds)
@@ -583,4 +635,4 @@ class GoogleDrive(DriveFolder):
                                 fileId=id_,
                                 fields=self.default_fields
                             ).execute()
-        return self._reply_to_object(result)
\ No newline at end of file
+        return self._reply_to_object(result)
